"""Example of using pydantic_cli leveraging the Pydantic Field
"""
from typing import Optional
from pydantic import BaseModel, Field

from pydantic_cli.examples import ConfigDefaults
from pydantic_cli import run_and_exit


class Options(BaseModel):

    class Config(ConfigDefaults):
        pass

    input_file: str = Field(
        ...,
        title="Input File",
        description="Path to the input file",
        # required=True, # this is implicitly set by ...
        extras={'cli': ("-f", "--input-file")}
    )

    max_records: int = Field(
        123,
        title="Max Records",
        description="Max number of records",
        gt=0,
        extras={'cli': ('-m', )},
    )

    min_filter_score: float = Field(
        ...,
        title="Min Score",
        description="Minimum Score Filter that will be applied to the records",
        extras={'cli': ('-s', )},
        gt=0
        # or extras={'cli': ('-s', '--min-filter-score', )}
    )

    max_filter_score: Optional[float] = Field(
        None,
        title="Max Score",
        description="Maximum Score Filter that will be applied to the records",
        gt=0,
        extras={'cli': ('-S', )}
        # or extras={'cli': ('-S', '--min-filter-score', )}
    )


def example_runner(opts: Options) -> int:
    print(f"Mock example running with options {opts}")
    print((opts.input_file, type(opts.input_file)))
    print(opts.max_records, type(opts.max_records))
    print(opts.min_filter_score, type(opts.min_filter_score))
    print(f"Fields {opts.fields}")
    return 0


if __name__ == "__main__":
    run_and_exit(Options, example_runner, description=__doc__, version='0.1.0')


"""
Example using Pydantic-CLI to generate a custom CLI fields on a per field basis.
"""
import sys
import logging

from pydantic import BaseModel

from pydantic_cli import __version__ as Version
from pydantic_cli import run_and_exit
from pydantic_cli.examples import ConfigDefaults

log = logging.getLogger(__name__)


class Options(BaseModel):

    class Config(ConfigDefaults):
        CLI_EXTRA_OPTIONS = {'max_records': ('-m', )}

    input_file: str
    max_records: int = 10


def example_runner(opts: Options) -> int:
    logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)
    log.info(f"pydantic_cli version={Version} Mock example running with options {opts}")
    return 0


if __name__ == "__main__":
    run_and_exit(Options, example_runner, description="Description", version='0.1.0')



"""
Example of Using a Subparser
"""
import typing as T
from pydantic import BaseModel, AnyUrl


from pydantic_cli.examples import ConfigDefaults
from pydantic_cli import run_sp_and_exit, SubParser


class AlphaOptions(BaseModel):

    class Config(ConfigDefaults):
        CLI_EXTRA_OPTIONS = {'max_records': ('-m', '--max-records')}

    input_file: str
    max_records: int = 10


class BetaOptions(BaseModel):

    class Config(ConfigDefaults):
        CLI_EXTRA_OPTIONS = {'url': ('-u', '--url'),
                             'num_retries': ('-n', '--num-retries')}

    url: AnyUrl
    num_retries: int = 3


def printer_runner(opts: T.Any):
    print(f"Mock example running with {opts}")
    return 0


def to_runner(sx):
    def example_runner(opts) -> int:
        print(f"Mock {sx} example running with {opts}")
        return 0
    return example_runner


def to_subparser_example():

    return {
        'alpha': SubParser(AlphaOptions, to_runner("Alpha"), "Alpha SP Description"),
        'beta': SubParser(BetaOptions, to_runner("Beta"), "Beta SP Description")}


if __name__ == "__main__":
    run_sp_and_exit(to_subparser_example(), description=__doc__, version='0.1.0')


import datetime
import sys
import traceback
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import logging
import typing as T


from pydantic import BaseModel
from pydantic.fields import ModelField

log = logging.getLogger(__name__)

NOT_PROVIDED = object()

VERSION = (1, 0, 0)

__version__ = ".".join([str(i) for i in VERSION])

__all__ = ['to_runner', 'run_and_exit',
           'to_runner_sp', 'run_sp_and_exit',
           'default_exception_handler',
           'default_prologue_handler',
           'default_epilogue_handler']


class SubParser(T.NamedTuple):
    options: T.Any
    runner_func: T.Callable
    description: T.Optional[T.AnyStr]


def _parser_add_version(parser: ArgumentParser, version: T.AnyStr) -> ArgumentParser:
    parser.add_argument("--version", action="version", version=version)
    return parser


def __to_field_description(default_value=NOT_PROVIDED, field_type=NOT_PROVIDED, description=None):
    desc = "" if description is None else description
    t = "" if field_type is NOT_PROVIDED else f"type:{field_type}"
    v = "" if default_value is NOT_PROVIDED else f"default:{default_value}"
    if not (t + v):
        xs = "".join([t, v])
    else:
        xs = " ".join([t, v])
    return f"{desc} ({xs})"


def __process_tuple(tuple_one_or_two, long_arg):
    """
    If the custom args are provided as only short, then
    add the long version.
    """
    lx = list(tuple_one_or_two)

    def is_short(xs):
        # xs = '-s'
        return len(xs) == 2

    if len(lx) == 1:
        first = lx[0]
        if is_short(first):
            return first, long_arg
        else:
            return tuple_one_or_two
    else:
        return tuple_one_or_two


def _add_pydantic_field_to_parser(p, ix, field: ModelField, override_cli=None) -> ArgumentParser:

    description = field.field_info.description
    extra = field.field_info.extra
    default_long_arg = f"--{ix}"

    is_positional = field.required

    # Should be a tuple2[Str, Str]
    try:
        cli_custom = __process_tuple(extra['extras']['cli'], default_long_arg)
        is_positional = False
    except KeyError:
        if override_cli is None:
            if is_positional:
                cli_custom = (ix, )
            else:
                cli_custom = (default_long_arg, )
        else:
            cli_custom = __process_tuple(override_cli, default_long_arg)

    log.debug(f"Creating Argument with opts:{cli_custom}, default={field.default} type={field.type_} positional={is_positional} required={field.required}")

    f = __to_field_description
    # this API is so thorny to code around. if dest='x' and p.add_argument('x') will raise
    if is_positional:
        p.add_argument(*cli_custom, help=f(NOT_PROVIDED, field.type_, description), default=field.default)
    else:
        p.add_argument(*cli_custom, help=f(field.default, field.type_, description), default=field.default, dest=ix, required=field.required)
    return p


def _add_pydantic_class_to_parser(p, cls) -> ArgumentParser:

    for ix, field in cls.__fields__.items():

        try:
            default_cli_opts = cls.Config.CLI_EXTRA_OPTIONS[ix]
        except (AttributeError, KeyError):
            default_cli_opts = None

        _add_pydantic_field_to_parser(p, ix, field,
                                      override_cli=default_cli_opts)

    return p


def pydantic_class_to_parser(cls, description=None, version=None) -> ArgumentParser:
    """
    Convert a pydantic data model class to an argparse instance
    """
    # Is there really not a lib for a JsonSchema Property to argparse option?

    p = ArgumentParser(description=description)

    _add_pydantic_class_to_parser(p, cls)

    if version is not None:
        _parser_add_version(p, version)

    return p


def default_exception_handler(ex) -> int:
    # this might need the opts instance, however
    # this isn't really well defined if there's an
    # error at that level
    sys.stderr.write(str(ex))
    exc_type, exc_value, exc_traceback = sys.exc_info()
    traceback.print_tb(exc_traceback, file=sys.stderr)
    return 1


def default_epilogue_handler(exit_code: int, run_time_sec: float):
    """
    General Hook to write that will be executed after the program is
    completed running
    """
    pass


def default_prologue_handler(opts) -> None:
    """
    General Hook to call before executing your runner func (e.g., f(opt)).

    Note this is semantically different from what argparse defines as a "prologue".

    This can be used to setup logging.

    :param opts: Will be an instance of your option class
    :return: None
    """
    pass


def _runner(args, parser: ArgumentParser, exception_handler, prologue_handler, epilogue_handler) -> int:

    def now():
        return datetime.datetime.now()

    started_at = now()
    try:
        # this is a really lackluster design
        # any --version or --help has sys.exit call
        # that is triggered from parser.exit()
        pargs = parser.parse_args(args)

        # This hackery is required for subparser case
        if not hasattr(pargs, 'cls'):
            parser.print_usage()
            parser.exit(1)

        cls = pargs.cls
        runner_func = pargs.func

        # to class instance from parsed arguments
        # is there a better way to do this?
        # log.debug(pargs.__dict__)
        opts = cls(**pargs.__dict__)

        # this validation interface is a bit odd
        cls.validate(opts)
        prologue_handler(opts)
        exit_code = runner_func(opts)
    except Exception as e:
        exit_code = exception_handler(e)

    dt = now() - started_at
    epilogue_handler(exit_code, dt.total_seconds())
    return exit_code


def _runner_with_args(
    args,
    cls,
    runner_func,
    description=None,
    version=None,
    exception_handler=default_exception_handler,
    prologue_handler=default_prologue_handler,
    epilogue_handler=default_epilogue_handler,
) -> int:

    # errors at the argparse level aren't always
    # communicated in an obvious way at this level
    parser = pydantic_class_to_parser(cls, description=description,
                                      version=version)

    # this is a bit of hackery
    parser.set_defaults(func=runner_func, cls=cls)

    return _runner(args, parser, exception_handler, prologue_handler, epilogue_handler)


def to_runner(
    cls,
    runner_func,
    description=None,
    version=None,
    exception_handler=default_exception_handler,
    prologue_handler=default_prologue_handler,
    epilogue_handler=default_epilogue_handler,
):
    def f(args):
        return _runner_with_args(
            args,
            cls,
            runner_func,
            description=description,
            version=version,
            exception_handler=exception_handler,
            prologue_handler=prologue_handler,
            epilogue_handler=epilogue_handler,
        )

    return f


def run_and_exit(
    cls,
    runner_func,
    description=None,
    version=None,
    exception_handler=default_exception_handler,
    prologue_handler=default_prologue_handler,
    epilogue_handler=default_epilogue_handler,
    args=sys.argv[1:],
):
    sys.exit(
        to_runner(
            cls,
            runner_func,
            description=description,
            version=version,
            exception_handler=exception_handler,
            prologue_handler=prologue_handler,
            epilogue_handler=epilogue_handler,
        )(args)
    )


def to_subparser(models: T.Dict[T.AnyStr, SubParser], description=None, version=None) -> ArgumentParser:
    p = ArgumentParser(description=description, formatter_class=ArgumentDefaultsHelpFormatter)

    sp = p.add_subparsers(help='commands')

    for subparser_id, sx in models.items():
        log.debug(("Adding Subparser", subparser_id, sx))

        spx = sp.add_parser(subparser_id, help=sx.description)
        _add_pydantic_class_to_parser(spx, sx.options)
        spx.set_defaults(func=sx.runner_func, cls=sx.options)

    if version is not None:
        _parser_add_version(p, version)

    return p


def to_runner_sp(subparsers: T.Dict[T.AnyStr, T.Type[BaseModel]],
                 description=None, version=None,
                 exception_handler=default_exception_handler,
                 prologue_handler=default_prologue_handler,
                 epilogue_handler=default_epilogue_handler) -> T.Callable[[T.List[T.AnyStr]], int]:

    def f(args):
        p = to_subparser(subparsers, description=description, version=version)

        return _runner(args, p, exception_handler, prologue_handler, epilogue_handler)

    return f


def run_sp_and_exit(subparsers: T.Dict[T.AnyStr, T.Type[BaseModel]],
                    description: T.Optional[T.AnyStr] = None,
                    version: T.Optional[T.AnyStr] = None,
                    exception_handler=default_exception_handler,
                    prologue_handler=default_prologue_handler,
                    epilogue_handler=default_epilogue_handler,
                    args=sys.argv[1:]):

    f = to_runner_sp(subparsers, description=description, version=version,
                     exception_handler=exception_handler,
                     prologue_handler=default_prologue_handler,
                     epilogue_handler=epilogue_handler)

    sys.exit(f(args))



"""
Example using Pydantic-CLI to generate a custom CLI fields on a per field basis
and setting up logging using the `prologue_handler`.

The `prologue_handler` provides a general mechanism to call
a setup hook with an instance of the Pydantic data model as the
argument to the function.
"""
import sys
from enum import Enum
import logging

from pydantic import BaseModel

from pydantic_cli import __version__
from pydantic_cli import run_and_exit
from pydantic_cli.examples import ConfigDefaults

log = logging.getLogger(__name__)


class LogLevel(str, Enum):
    # wish this was defined in the stdlib's logging as an enum
    INFO = "INFO"
    DEBUG = "DEBUG"
    WARN = "WARN"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class Options(BaseModel):

    class Config(ConfigDefaults):
        CLI_EXTRA_OPTIONS = {'max_records': ('-m', ),
                             'log_level': ('-l', )}

    input_file: str
    max_records: int = 10
    # this leverages Pydantic's fundamental understanding of Enums
    log_level: LogLevel = LogLevel.INFO


def prologue_handler(opts: Options):
    """Define a general Prologue hook to setup logging for the application"""
    format_str = '[%(levelname)s] %(asctime)s [%(name)s %(funcName)s %(lineno)d] %(message)s'
    logging.basicConfig(level=opts.log_level.upper(), stream=sys.stdout, format=format_str)
    log.info(f"Set up log with level {opts.log_level}")
    log.info(f"Running {__file__}")


def epilogue_handler(exit_code: int, run_time_sec: float) -> None:
    log.info(f"Completed running {__file__} with exit code {exit_code} in {run_time_sec} sec.")


def example_runner(opts: Options) -> int:
    log.info(f"pydantic_cli version={__version__} Mock example running with options {opts}")
    return 0


if __name__ == "__main__":
    run_and_exit(Options, example_runner, description="Description", version='0.1.0',
                 prologue_handler=prologue_handler,
                 epilogue_handler=epilogue_handler)



import os
from setuptools import setup

version = __import__('pydantic_cli').__version__


def _get_local_file(file_name):
    return os.path.join(os.path.dirname(__file__), file_name)


def _read(file_name):
    with open(_get_local_file(file_name)) as f:
        return f.read()


setup(name='pydantic_cli',
      version=version,
      description='Turn Pydantic defined Data Models into CLI Tools',
      long_description=_read('README.md'),
      long_description_content_type="text/markdown",
      url='http://github.com/mpkocher/pydantic-cli',
      author='M. Kocher',
      author_email='michael.kocher@me.com',
      license='MIT',
      packages=['pydantic_cli', 'pydantic_cli.examples'],
      tests_require = ['nose'],
      zip_safe=False,
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
      ]
      )



"""
Simple Example of using Pydantic to generate a CLI

This example will generate a CLI tool with 1 positional argument and one optional boolean arg
"""
from pydantic import BaseModel

from pydantic_cli import run_and_exit


class Options(BaseModel):
    input_file: str
    run_training: bool = False


def example_runner(opts: Options) -> int:
    print(f"Mock example running with {opts}")
    print(f"Fields {opts.fields}")
    return 0


if __name__ == "__main__":
    run_and_exit(Options, example_runner, description=__doc__, version='0.1.0')
